/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * config.c - config file reader
 * Copyright (c) 2020-2021, 2023 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <fcl/types.h>
#include <fcl/string.h>
#define FCL_SIMPLE_NAMES
#include <fcl/strparse.h>
#include <fcl/malloc.h>
#include <fcl/qprint.h>
#include <libsimpletcp/libsimpletcp.h>
#include "eventlog.h"
#include "helpers/conn_open.h"
#include "config.h"

typedef struct {
  char const * name;
  int          type;
  union {
    uchar        * _V_uchar;
    uint         * _V_uint;
    ulong        * _V_ulong;
    uint16       * _V_uint16;
    uint32       * _V_uint32;
    uint64       * _V_uint64;
    size_t       * _V_size_t;
    char const * * _V_string;
  } V;
  union {
    uchar        _D_uchar;
    uint         _D_uint;
    ulong        _D_ulong;
    uint16       _D_uint16;
    uint32       _D_uint32;
    uint64       _D_uint64;
    size_t       _D_size_t;
    char const * _D_string;
  } D;
} t_config_item;

#define _T_uchar      1
#define _T_uint       2
#define _T_ulong      3
#define _T_uint16     4
#define _T_uint32     5
#define _T_uint64     6
#define _T_size_t     7
#define _T_string     8

#define PAR(_type, _varname, _default)  { .name = #_varname, .type = _T_##_type, .V._V_##_type = &prefs._varname, .D._D_##_type = _default }
#define PARH(_type, _varname, _default)  { .name = #_varname, .type = _T_##_type | 256, .V._V_##_type = &prefs._varname, .D._D_##_type = _default }

t_prefs prefs;

static t_config_item config[] = {
  /* type           varname                         default    */
  PAR(string,       base_dir,                       "."                                  ),
  PARH(uint32,      listen.ip,                      0x7F00000A                           ),
  PAR(uint16,       listen.port,                    3128                                 ),
  PAR(uchar,        listen.raw_https,               0                                    ),
  PAR(uint16,       num_processes,                  40                                   ),
  PAR(uint,         connection_timeout.keepalive,   30                                   ),
  PAR(uint,         connection_timeout.idle,        36000                                ),
  PARH(uint16,      local_ssl.minver,               0x0303                               ),
  PAR(string,       local_ssl.cacert,               "./pem/cacert.pem"                   ),
  PAR(string,       local_ssl.cakey,                "./pem/cakey.pem"                    ),
  PAR(string,       local_ssl.proxykey,             "./pem/proxykey.pem"                 ),
  PAR(uchar,        local_ssl.wildcard_certs,       1                                    ),
  PAR(uchar,        local_ssl.no_SAN,               0                                    ),
  PARH(uint32,      rproxy.host_dns,                0x7F00000A                           ),
  PAR(uint16,       rproxy.port_dns,                1115                                 ),
  PAR(string,       helpers.ssl_server_wrapper,     "./ssl-server-wrapper"               ),
  PAR(string,       client.wrapper_https,           "./fproxy-client-wrapper"            ),
  PAR(string,       client.wrapper_http,            ""                                   ),
  PAR(string,       client.proxy,                   "r://127.0.0.10:1116"                ),
  PAR(uchar,        client.always_use_proxy,        0                                    ),
  PAR(uchar,        client.use_wrapper_connect,     0                                    ),
  PARH(uint16,      client.ssl_minver,              0x0301                               ),
  PAR(string,       client.ssl_calist,              "/etc/ssl/certs/ca-certificates.crt" ),
  { NULL }
};

static void config_check(void);
int config_load(char const * fn) {
  FILE * fp;
  t_config_item * c;
  char buf[1000];
  char ch, *p, *name;
  unsigned int ln;
  uint32 tmp32;
  
  bzero(&prefs, sizeof(prefs));
  ln = 0;
  for(c=config; c->name; c++) switch(c->type & 255) {
  default: goto unkt;
#define DD(type) case _T_##type: *c->V._V_##type = c->D._D_##type; break;
  DD(uchar)
  DD(uint)
  DD(ulong)
  DD(uint16)
  DD(uint32)
  DD(uint64)
  DD(string)
  DD(size_t)
#undef DD
  }
  
  if(!fn) return 0;
  if(!(fp=fopen(fn,"rb"))) { eventlog(eventlog_level_error, __FUNCTION__, "open config file \"%s\" error %d (%s)", fn, errno, strerror(errno)); exit(-1); }
  while(fgets(buf, sizeof(buf), fp)) {
    ln++;
    if(!(p=strchr(buf,'\n')) || p[1]) goto nt;
    *p = 0;
    p = (char*)skip_spaces(buf);
    if(!*p || *p=='#') continue;
    for(name=p; (ch=*p) && ch!=' ' && ch!='\t' && ch!='='; p++);
    if(!ch) goto bad;
    *(p++) = 0;
    if(ch!='=') {
      p = (char*)skip_spaces(p);
      if(*p!='=') goto bad;
      p++;
    }
    p = (char*)skip_spaces(p);
    for(c=config; c->name && strcmp(c->name,name); c++);
    if(!c->name) goto unk;
    switch(c->type) {
    default: goto unkt;
    case _T_uchar: if(str_to_uchar(p,c->V._V_uchar)<0) goto badv; break;
    case _T_uint: if(str_to_uint(p,c->V._V_uint)<0) goto badv; break;
    case _T_ulong: if(str_to_ulong(p,c->V._V_ulong)<0) goto badv; break;
    case _T_uint16: if(str_to_uint16(p,c->V._V_uint16)<0) goto badv; break;
    case _T_uint16|256: if(str_hex_to_uint16(p,c->V._V_uint16)<0) goto badv; break;
    case _T_uint32: if(str_to_uint32(p,c->V._V_uint32)<0) goto badv; break;
    case _T_uint32|256: if((tmp32=lst_parse_ipstr(p))==LST_INADDR_NONE) goto badv; *c->V._V_uint32 = tmp32; break;
    case _T_uint64: if(str_to_uint64(p,c->V._V_uint64)<0) goto badv; break;
    case _T_string: *c->V._V_string = fcl_strdup_check(p); break;
    }
  }
  config_check();
  return 0;
nt: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u non-terminated line or line too long", fn, ln); exit(-1);
bad: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u syntax error", fn, ln); exit(-1);
unk: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u unknown option %s", fn, ln, name); exit(-1);
unkt: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u unknown type for option %s", fn, ln, c->name); exit(-1);
badv: eventlog(eventlog_level_error, __FUNCTION__, "config file \"%s\":%u bad value for option %s", fn, ln, c->name); exit(-1);
}

static void fix_binpath(char const *title, char const **value, size_t minlen, int allow_addr) {
  char const *v, *p;
  char tmp[1000];
  size_t len, len2, j;
  if(!value || !(v=*value) || !(len=strlen(v))) return;
  if(allow_addr && (skip_prefix(v,"tcp:") || skip_prefix(v,"unix:"))) return;
  if(len>400) { eventlog(eventlog_level_error,__FUNCTION__, "param \"%s\" value too long: \"%s\"", title, v); exit(-1); }
  if(minlen>400) minlen = 400;
  if(p=strrchr(v,'/')) {
    if(access(v,X_OK)) { eventlog(eventlog_level_error,__FUNCTION__, "param \"%s\" value \"%s\" doesn't match any executable file", title, v); exit(-1); }
    p++;
    if(!*p) { eventlog(eventlog_level_error,__FUNCTION__, "param \"%s\" value \"%s\" doesn't match any executable file", title, v); exit(-1); }
    if(len>=minlen) return;
    j = p-v;
    bcopy(v, tmp, j);
    len = strlen(p);
    assert(len<minlen && minlen-len>j);
    while(j<minlen-len && j<400) { tmp[j++] = '.'; tmp[j++] = '/'; }
    bcopy(p, tmp+j, len);
    tmp[j+len] = 0;
  } else {
    sprintf(tmp, "/usr/local/bin/%s", v);
    j = 15;
    if(access(tmp,X_OK)) {
      sprintf(tmp, "/usr/bin/%s", v);
      j = 9;
      if(access(tmp,X_OK)) { eventlog(eventlog_level_error,__FUNCTION__, "param \"%s\" value \"%s\" doesn't match any executable file in /usr/local/bin or /usr/bin", title, v); exit(-1); }
    }
    len2 = strlen(tmp);
    if(len2<minlen) {
      assert(len<minlen);
      while(j<minlen-len && j<400) { tmp[j++] = '.'; tmp[j++] = '/'; }
      bcopy(v, tmp+j, len);
      tmp[j+len] = 0;
    }
  }
  *value = fcl_strdup_check(tmp);
}

static void config_check(void) {
  int r;
  static proxy_desc proxy_parsed;
  fix_binpath("client.wrapper_https", &prefs.client.wrapper_https, 70, 1);
  fix_binpath("client.wrapper_http", &prefs.client.wrapper_http, 70, 1);
  fix_binpath("helpers.ssl_server_wrapper", &prefs.helpers.ssl_server_wrapper, 0, 0);
  if(prefs.client.proxy && *prefs.client.proxy) {
    r = parse_proxyspec(&proxy_parsed, prefs.client.proxy);
    if(r==-2) { eventlog(eventlog_level_error,__FUNCTION__, "out of memory parsing \"client.proxy\" value \"%s\"", prefs.client.proxy); exit(-1); }
    if(r<0) { eventlog(eventlog_level_error,__FUNCTION__, "error parsing \"client.proxy\" value \"%s\"", prefs.client.proxy); exit(-1); }
    prefs.client.proxy_parsed = &proxy_parsed;
  }
}

void config_dump(void) {
  t_config_item * c;
  uint32 n, *d;
  char buf[50];
  
  for(c=config; c->name; c++) switch(c->type) {
#define DD(type, t2, fmt) case _T_##type: qp_printf(&qp_out, " %10s %-20s = "fmt"\n", #type, c->name, t2 (*c->V._V_##type)); break;
#define DH(type, t2, fmt) case _T_##type|256: qp_printf(&qp_out, " %10s %-20s = "fmt"\n", #type, c->name, t2 (*c->V._V_##type)); break;
  DD(uchar, (uint), "%u")
  DD(uint, , "%u")
  DD(ulong, , "%lu")
  DD(uint16, (uint), "%u")
  DH(uint16, (uint), "%04X")
  DD(uint32, (ulong), "%lu")
  DH(uint32, (uint32), "%{IP32}")
  DD(uint64, (ullong), "%llu")
  DD(string, , "%s")
  DD(size_t, (ullong), "%llu")
#undef DD
#undef DH
  }
  qp_flush(&qp_out);
}



