/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * handle/rules.resolv.c - rules_resolv.txt parse and lookup
 * Copyright (c) 2018, 2023 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <assert.h>
#include <fcl/types.h>
#include <fcl/malloc.h>
#include <fcl/strparse.h>
#include <libsimpletcp/libsimpletcp.h>
#include "../config.h"
#include "../util.h"
#include "../eventlog.h"
#include "../helpers/nslookup.h"
#include "../helpers/conn_open.h"
#include "rules.route.h"

/*
# format
# hostspec hostspec hostspec
# =>\tconnspec
# hostspec is [http://|https://]hostname[:port][/prefix]
# connspec is [http://|https://]ip[:port][/prefix]
# unspecified protocol in hostspec = match both http/https
# unspecified protocol in connspec = keep original protocol
# unspecified port in hostspec = match any port
# unspecified port in connspec = (allowed only if protocol also unspecified) keep original port
# /prefix is unsupported for now

Новый:
pattern[ pattern pattern ...]
	cmd
	cmd
	cmd
	cmd

	resolve off
	resolve <domain> [via <addr>]
	set target <target>
	set port <port>
	set proto <proto>
	set wrapper <wrapper>
	set proxy <proxy_spec>
	set use_wrapper_connect <value>

*/

#define _MAX_HOSTS 16
#define _MAX_DOMAINS 10000

typedef struct {
  char const  *name;   /* can't be NULL */
  char const  *prefix; /* NULL = unspecified */
  uint16      port;    /* 0=any */
  char        proto;   /* 0=any, 1=http, 2=https */
} t_rules_host;
typedef struct {
  unsigned int n_hosts;
  t_rules_host *hosts[_MAX_HOSTS];
  char         disable_resolve;     /* 0=do not disable; 1=disable */
  char         use_wrapper_connect; /* 0=default, 1=disable, 2=enable */
  char const   *new_resolve_domain; /* NULL=do not alter domain */
  uint32       resolve_via_addr;    /* if port==0: 0=default, -1=force non-proxy */
  uint16       resolve_via_port;    /* 0=special (see addr) */
  uint32       new_addr;
  uint16       new_port;
  char         new_proto;           /* 0=don't change, 1=http, 2=https */
  char const   *new_host_port;
  char const   *new_host;
  char const   *new_prefix;
  char const   *new_wrapper;
  proxy_desc   proxy;
  char         set_proxy;
  char         do_continue;
} t_rules_domain;
typedef struct {
  unsigned int    n, a;
  t_rules_domain  **domains;
} t_rules_list;

static t_rules_domain * rules_add_domain(t_rules_list *r) {
  t_rules_domain  *d;
  unsigned int    n, inc;
  t_rules_domain  **domains;
  n = r->n;
  if(n>=_MAX_DOMAINS) return NULL;
  if(n==r->a) {
    inc = 20 + r->a/3;
    if(_MAX_DOMAINS-r->a>inc) inc+=r->a; else inc=_MAX_DOMAINS;
    domains = fcl_realloc_a_check(r->domains, sizeof(*domains), inc);
    r->domains = domains;
    r->a = inc;
  } else domains = r->domains;
  assert(n<r->a);
  d = fcl_mallocz_check(sizeof(*d));
  domains[n] = d;
  r->n = n+1;
  return d;
}

/* -1=too many hosts; -2=syntax error */
static int domain_add_host(t_rules_domain *d, char const *hostspec) {
  unsigned int  n;
  t_rules_host  h;
  char          *hs, *p0, *p;
  n = d->n_hosts;
  if(n>=_MAX_HOSTS) return -1;
  bzero(&h, sizeof(h));
  p0 = hs = fcl_strdup_check(hostspec);
  if(p=strstr(hs,"://")) {
    *p = 0;
    if(!strcmp(hs,"http")) h.proto = 1;
    else if(!strcmp(hs,"https")) h.proto = 2;
    else { fcl_free(p0); return -2; }
    hs = p+3;
  }
  if(p=strchr(hs,'/')) {
    h.prefix = fcl_strdup_check(p);
    *p = 0;
  }
  if(p=strchr(hs,':')) {
    *p = 0;
    if(str_to_uint16(p+1, &h.port)<0 || !h.port) { fcl_free(p0); fcl_free_c(h.prefix); return -2; }
  }
  h.name = fcl_strdup_check(hs);
  free(p0);
  d->hosts[n] = fcl_memdup_check(&h, sizeof(h));
  d->n_hosts = n+1;
  return 0;
}

/* -1=error, already logged */
static int domain_parse_cmd(t_rules_domain *d, char *cmd, char const *fn, unsigned int ln) {
#define NEXTARG for(cmd=p; *p; p++) if(*p==' ') { *(p++) = 0; while(*p==' ') p++; break; }
  char *p, *pp;
  size_t len;
  while(*cmd==' ') cmd++;
  for(p=cmd; *p; p++) if(*p==' ') { *(p++) = 0; while(*p==' ') p++; break; }
  if(!strcmp(cmd,"resolve")) {
    NEXTARG
    if(!strcmp(cmd,"off")) {
      if(*p) { eventlog(eventlog_level_error,__FUNCTION__,"extra args for \"resolve off\" at \"%s\":%u", fn, ln); return -1; }
      d->disable_resolve = 1;
      return 0;
    }
    if(*cmd=='"') {
      len = strlen(cmd);
      if(len<3 || cmd[len-1]!='"') { eventlog(eventlog_level_error,__FUNCTION__,"quote syntax error at \"%s\":%u", fn, ln); return -1; }
      if(d->new_resolve_domain || d->new_addr) { eventlog(eventlog_level_error,__FUNCTION__,"only one of \"resolve <domain>\" and \"set target\" could be specified at \"%s\":%u", fn, ln); return -1; }
      d->new_resolve_domain = fcl_substrdup_check(cmd+1, len-2);
      NEXTARG
    }
    if(!strcmp(cmd,"via")) {
      NEXTARG
      if(*p) { eventlog(eventlog_level_error,__FUNCTION__,"extra args for \"resolve via\" at \"%s\":%u", fn, ln); return -1; }
      if(!strcmp(cmd,"direct")) { d->resolve_via_addr = 0xFFFFFFFF; d->resolve_via_port = 0; }
      else if(lst_parse_addrstr(cmd,&d->resolve_via_addr,&d->resolve_via_port)!=3 || !d->resolve_via_port) { eventlog(eventlog_level_error,__FUNCTION__,"bad proxy_r address for \"resolve via\" at \"%s\":%u", fn, ln); return -1; }
      return 0;
    }
    if(!*cmd) return 0;
    eventlog(eventlog_level_error,__FUNCTION__,"extra/unknown args for \"resolve\" at \"%s\":%u", fn, ln); return -1;
  }
  if(!strcmp(cmd,"continue")) {
    if(*p) { eventlog(eventlog_level_error,__FUNCTION__,"extra args for \"continue\" at \"%s\":%u", fn, ln); return -1; }
    d->do_continue = 1;
    return 0;
  }
  if(strcmp(cmd,"set")) { eventlog(eventlog_level_error,__FUNCTION__,"unknown command at \"%s\":%u", fn, ln); return -1; }
  if(!*p) { eventlog(eventlog_level_error,__FUNCTION__,"missing args for \"set\" command at \"%s\":%u", fn, ln); return -1; }
  NEXTARG
  if(!*p && strcmp(cmd,"proxy")) { eventlog(eventlog_level_error,__FUNCTION__,"missing second arg for \"set\" command at \"%s\":%u", fn, ln); return -1; }
  if(strchr(p,' ')) { eventlog(eventlog_level_error,__FUNCTION__,"extra args for \"set\" command at \"%s\":%u", fn, ln); return -1; }
  if(!strcmp(cmd,"port")) {
    if(d->new_port) { eventlog(eventlog_level_error,__FUNCTION__,"new port already specified at \"%s\":%u", fn, ln); return -1; }
    if(str_to_uint16(p,&d->new_port)<0 || !d->new_port) { eventlog(eventlog_level_error,__FUNCTION__,"bad value for \"set port\" at \"%s\":%u", fn, ln); return -1; }
    return 0;
  }
  if(!strcmp(cmd,"proto")) {
    if(d->new_proto) { eventlog(eventlog_level_error,__FUNCTION__,"new proto already specified at \"%s\":%u", fn, ln); return -1; }
    if(!strcmp(p,"http")) d->new_proto = 1;
    else if(!strcmp(p,"https")) d->new_proto = 2;
    else { eventlog(eventlog_level_error,__FUNCTION__,"bad value for \"set proto\" at \"%s\":%u", fn, ln); return -1; }
    return 0;
  }
  if(!strcmp(cmd,"prefix")) {
    if(d->new_prefix) { eventlog(eventlog_level_error,__FUNCTION__,"new prefix already specified at \"%s\":%u", fn, ln); return -1; }
    d->new_prefix = fcl_strdup_check(p);
    return 0;
  }
  if(!strcmp(cmd,"http_host")) {
    if(d->new_host_port) { eventlog(eventlog_level_error,__FUNCTION__,"new http_host already specified at \"%s\":%u", fn, ln); return -1; }
    d->new_host_port = fcl_strdup_check(p);
    if(pp=strchr(p,':')) *pp = 0;
    d->new_host = fcl_strdup_check(p);
    return 0;
  }
  if(!strcmp(cmd,"wrapper")) {
    if(d->new_wrapper) { eventlog(eventlog_level_error,__FUNCTION__,"new wrapper already specified at \"%s\":%u", fn, ln); return -1; }
    d->new_wrapper = fcl_strdup_check(p);
    return 0;
  }
  if(!strcmp(cmd,"use_wrapper_connect")) {
    switch(str_get_bool(p)) {
    case 0: d->use_wrapper_connect = 0; return 0;
    case 1: d->use_wrapper_connect = 1; return 0;
    }
    eventlog(eventlog_level_error,__FUNCTION__,"bad value for \"set use_wrapper_connect\" at \"%s\":%u", fn, ln); return -1;
  }
  if(!strcmp(cmd,"proxy")) {
    if(d->set_proxy) { eventlog(eventlog_level_error,__FUNCTION__,"new proxy already specified at \"%s\":%u", fn, ln); return -1; }
    if(!*p) { d->set_proxy = 2; return 0; }
    d->set_proxy = 1;
    switch(parse_proxyspec(&d->proxy,p)) {
    case 0: return 0;
    case -2: eventlog(eventlog_level_error,__FUNCTION__, "out of memory parsing \"set proxy\" at \"%s\":%u", fn, ln); return -1;
    default: eventlog(eventlog_level_error,__FUNCTION__, "bad value for \"set proxy\" at \"%s\":%u", fn, ln); return -1;
    }
  }
  if(strcmp(cmd,"target")) { eventlog(eventlog_level_error,__FUNCTION__,"unknown \"set\" subcommand at \"%s\":%u", fn, ln); return -1; }
  if(d->new_resolve_domain || d->new_addr) { eventlog(eventlog_level_error,__FUNCTION__,"only one of \"resolve <domain>\" and \"set target\" could be specified at \"%s\":%u", fn, ln); return -1; }
  d->disable_resolve = 1;
  if(pp=strstr(p,"://")) {
    if(d->new_proto) { eventlog(eventlog_level_error,__FUNCTION__,"new proto already specified at \"%s\":%u", fn, ln); return -1; }
    *pp = 0;
    if(!strcmp(p,"http")) d->new_proto = 1;
    else if(!strcmp(p,"https")) d->new_proto = 2;
    else { eventlog(eventlog_level_error,__FUNCTION__,"bad value for \"set target\" proto at \"%s\":%u", fn, ln); return -1; }
    p = pp+3;
  }
  if(pp=strchr(p,'/')) {
    if(d->new_prefix) { eventlog(eventlog_level_error,__FUNCTION__,"new prefix already specified at \"%s\":%u", fn, ln); return -1; }
    d->new_prefix = fcl_strdup_check(pp);
    *pp = 0;
  }
  if(pp=strchr(p,':')) {
    if(d->new_port) { eventlog(eventlog_level_error,__FUNCTION__,"new port already specified at \"%s\":%u", fn, ln); return -1; }
    if(str_to_uint16(pp+1,&d->new_port)<0 || !d->new_port) { eventlog(eventlog_level_error,__FUNCTION__,"bad value for \"set target\" port at \"%s\":%u", fn, ln); return -1; }
    *pp = 0;
  }
  if((d->new_addr=lst_parse_ipstr(p))==LST_INADDR_NONE) { d->new_addr=0; d->new_resolve_domain = fcl_strdup_check(p); }
  return 0;
}

static void rules_list_clear(t_rules_list *r) {
  unsigned int  n, n1, n2, n3;
  t_rules_domain *d;
  t_rules_domain **domains;
  t_rules_host *host;
  if(domains=r->domains) {
    for(n=r->n; n--; ) {
      d = domains[n];
      for(n1=d->n_hosts; n1--; ) {
        host = d->hosts[n1];
        fcl_free_c_check(host->name);
        fcl_free_c(host->prefix);
        fcl_free(host);
      }
      fcl_free_c(d->new_prefix);
      fcl_free_c(d->new_resolve_domain);
      fcl_free_c(d->new_wrapper);
      free_proxyspec(&d->proxy);
      fcl_free(d);
    }
    fcl_free(domains);
  }
  bzero(r, sizeof(*r));
}

static int rules_list_load(t_rules_list *r, char const *fn) {
  FILE      *fp;
  int       rr;
  char      buf[1024];
  char      *url, *p, *p2;
  size_t    len;
  unsigned int    ln, type;
  t_rules_domain  *d;
  
  d = NULL;
  bzero(r, sizeof(*r));
  if(!(fp=fopen(fn,"r"))) { eventlog(eventlog_level_error,__FUNCTION__,"error opening resolv-rules file \"%s\" for reading", fn); return -1; }
  ln = 0;
  while(fgets(buf, sizeof(buf), fp)) {
    ln++;
    len = strlen(buf);
    if(!len || buf[len-1]!='\n') { eventlog(eventlog_level_error,__FUNCTION__,"line too long or missing end-of-line at \"%s\":%u", fn, ln); goto fail; }
    buf[--len] = 0;
    for(len=0; buf[len] && buf[len]!='#'; len++);
    while(len && buf[len-1]==' ') len--;
    buf[len] = 0;
    if(!len) continue;
    if(buf[0]=='\t') {
      if(!d) { eventlog(eventlog_level_error,__FUNCTION__,"missing hostspec in previous line at \"%s\":%u", fn, ln); goto fail; }
      if(domain_parse_cmd(d, buf+1, fn, ln)<0) goto fail;
      continue;
    }
    if(!(d=rules_add_domain(r))) { eventlog(eventlog_level_error,__FUNCTION__,"domain - too many domains in entire config at \"%s\":%u", fn, ln); goto fail; }
    p2 = buf;
    while(*p2) {
      if(p=strchr(p2,' ')) *(p++) = 0;
      rr = domain_add_host(d,p2);
      if(rr==-1) { eventlog(eventlog_level_error,__FUNCTION__,"domain - too many hostnames in line at \"%s\":%u", fn, ln); goto fail; }
      if(rr<0) { eventlog(eventlog_level_error,__FUNCTION__,"domain - syntax error at \"%s\":%u", fn, ln); goto fail; }
      if(!p) break;
      while(*p==' ') p++;
      p2 = p;
    }
  }
  fclose(fp);
  return 0;
fail:
  rules_list_clear(r);
  fclose(fp);
  return -1;
}

/* 1=matched 0=no */
static int host_match_sub(char const *w, char const *h) {
  size_t wlen, hlen;
  if(!*w) return 0;
  if(w[0]=='*' && !w[1]) return 1;
  if(!strcasecmp(w,h)) return 1;
  if(w[0]=='*' && w[1]=='.') {
    wlen = strlen(w);
    hlen = strlen(h);
    if(hlen>=wlen && !strcasecmp(w+1,h+(hlen-wlen+1))) return 1;
  }
  if(w[0]=='.') {
    wlen = strlen(w);
    hlen = strlen(h);
    if(hlen<wlen) return (wlen-hlen==1 && !strcasecmp(w+1,h));
    return !strcasecmp(w,h+(hlen-wlen));
  }
  return 0;
}

/* 1=matched 0=no */
static int host_match(t_rules_domain *d, char const *host, int is_https, uint16 port, char const *url, size_t *prefix_len) {
  int sub;
  unsigned int n, j;
  t_rules_host *h;
  
  n = d->n_hosts;
  for(j=0; j<n; j++) {
    h = d->hosts[j];
    if(h->proto && h->proto!=(is_https?2:1)
      || h->port && h->port!=port
      || h->prefix && (!url || !str_skip_prefix(url,h->prefix))) continue;
    if(host_match_sub(h->name, host)) { if(prefix_len) *prefix_len=h->prefix?strlen(h->prefix):0; return 1; }
  }
  return 0;
}

static t_rules_domain * search_route_rule(t_rules_list *r, char const *host, int is_https, uint16 port, char const *url, size_t *prefix_len) {
  t_rules_domain  *d;
  unsigned int    n, j;
  for(j=0,n=r->n; j<n; j++) if(host_match(d=r->domains[j], host, is_https, port, url, prefix_len)) return d;
  return NULL;
}



static t_rules_list list_route;
extern void route_rules_load(void) {
  if(rules_list_load(&list_route, "rules_route.txt")<0) {
    eventlog(eventlog_level_fatal,__FUNCTION__,"can't load ROUTE rules");
    exit(-1);
  }
}



extern int do_route(t_route_settings *rs) {
  /* we expect r->orig_* fields are filled and others are zeroed */
  t_rules_domain *d;
  size_t prefix_len, new_prefix_len, url_tail_len, new_url_len;
  uint32 dns_via_addr;
  uint16 dns_via_port;
  char *new_url;
  unsigned int n, j;
  char disable_resolve;
  rs->use_wrapper_connect = prefs.client.use_wrapper_connect;
  rs->https = rs->orig_https;
  rs->new_host_port = NULL;
  rs->new_host = NULL;
  rs->addr = 0;
  rs->port = rs->orig_port;
  rs->resolve_domain = NULL;
  rs->new_url = NULL;
  rs->wrapper = rs->https?prefs.client.wrapper_https:prefs.client.wrapper_http;
  rs->proxy = prefs.client.always_use_proxy?prefs.client.proxy_parsed:NULL;
  
  if(prefs.client.always_use_proxy) { dns_via_addr = prefs.rproxy.host_dns; dns_via_port = prefs.rproxy.port_dns; }
  else { dns_via_addr = 0; dns_via_port = 0; }
  disable_resolve = 0;

/*  if(d = search_route_rule(&list_route, rs->orig_host, rs->orig_https, rs->orig_port, rs->orig_url, &prefix_len)) { */
  for(j=0,n=list_route.n; j<n; j++) if(host_match(d=list_route.domains[j], rs->orig_host, rs->orig_https, rs->orig_port, rs->orig_url, &prefix_len)) {
    if(d->use_wrapper_connect) rs->use_wrapper_connect = d->use_wrapper_connect-1;
    if(d->new_resolve_domain) rs->resolve_domain = d->new_resolve_domain;
    if(d->new_addr) rs->addr = d->new_addr;
    if(d->new_host_port) rs->new_host_port = d->new_host_port;
    if(d->new_host) rs->new_host = d->new_host;
    if(d->new_port) rs->port = d->new_port;
    if(d->new_proto) rs->https = d->new_proto-1;
    if(d->new_wrapper) rs->wrapper = d->new_wrapper;
    if(d->set_proxy==1) rs->proxy = &d->proxy;
    if(d->set_proxy==2 && !rs->proxy) rs->proxy = prefs.client.proxy_parsed;
    if(d->new_prefix) {
      new_prefix_len = strlen(d->new_prefix);
      url_tail_len = strlen(rs->orig_url+prefix_len);
      new_url_len = new_prefix_len+url_tail_len+1;
      if(new_url_len<=url_tail_len) return -1;
      /* TODO: handle as out of memory?
               really this couldn't happen because rules parser has line length limit and url has some limit too */
      new_url = fcl_malloc_check(new_url_len);
      if(new_prefix_len) bcopy(d->new_prefix, new_url, new_prefix_len);
      if(url_tail_len) bcopy(rs->orig_url+prefix_len, new_url+new_prefix_len, url_tail_len);
      new_url[new_prefix_len+url_tail_len] = 0;
      if(rs->new_url) fcl_free_c_check(rs->new_url);
      rs->new_url = new_url;
    }
    if(d->resolve_via_port) { dns_via_addr = d->resolve_via_addr; dns_via_port = d->resolve_via_port; }
    else if(d->resolve_via_addr==0xFFFFFFFF) { dns_via_addr = 0; dns_via_port = 0; }
    if(d->disable_resolve) disable_resolve = 1;
    if(!d->do_continue) break;
  }
  if(!disable_resolve) {
    if(nslookup(rs->resolve_domain?rs->resolve_domain:rs->orig_host, &rs->addr, dns_via_addr, dns_via_port)<0) return -1;
  }
  return 0;
}





