/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/conn_open.c - open a connection through various paths
 * Copyright (c) 2023 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <fcl/types.h>
#include <fcl/endian.h>
#include <fcl/string.h>
#include <fcl/malloc.h>
#define FCL_SIMPLE_NAMES
#include <fcl/strparse.h>
#include <fcl/qprint.h>
#include <libsimpletcp/libsimpletcp.h>
#include "nslookup.h"
#include "conn_open.h"

/* return: -1=parse error; -2=out of memory */
extern int parse_proxyspec(proxy_desc *r, char const *b) {
  char const *a, *auth;
  proxy_type t;
  uint32 addr;
  uint16 port;
  
  if(!strcmp(b,"none")) { bzero(r,sizeof(*r)); return 0; }
  if(a=skip_prefix(b,"http://")) t = proxy_http;
  else if(a=skip_prefix(b,"https://")) t = proxy_https;
  else if(a=skip_prefix(b,"socks4://")) t = proxy_socks4;
  else if(a=skip_prefix(b,"socks5://")) t = proxy_socks5;
  else if(a=skip_prefix(b,"r://")) t = proxy_r;
  else return -1;
  if(b = strrchr(a,'@')) {
    if(lst_parse_addrstr(b+1,&addr,&port)!=3) return -1;
    if(!(auth = fcl_substrdup(a,b-a))) return -2;
  } else {
    if(lst_parse_addrstr(a,&addr,&port)!=3) return -1;
    auth = NULL;
  }
  r->type = t;
  r->ip = addr;
  r->port = port;
  r->auth = auth;
  return 0;
}

extern void free_proxyspec(proxy_desc *r) {
  if(r->auth) fcl_free_c_check(r->auth);
  bzero(r, sizeof(*r));
}

static int handle_proxy_http(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata);
static int handle_proxy_https(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata);
static int handle_proxy_socks4(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata);
static int handle_proxy_socks5(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata);
static int handle_proxy_r(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata);

/* TODO: LST_FLAG_NONBLOCK_BEFORE support (split the function into initial and process func, add ctx struct) */
#define ERR0(msg) do { if(logfn) logfn(cbdata, "%s", msg); } while(0)
#define ERR(fmt, ...) do { if(logfn) logfn(cbdata, fmt, __VA_ARGS__); } while(0)
extern int conn_open_via(uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, uint32 flags, conn_open_log_fn logfn, void *cbdata, int verbose) {
  int fd, r;
  int (*handler)(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata);
  char const *name;
  
  if(!via || via->type==proxy_none) {
    if(!addr) {
      if(!hostname) { ERR0("both target addr and target hostname missing"); return -1; }
      if(verbose) ERR("resolving hostname %s", hostname);
      if(nslookup(hostname, &addr, 0, 0)<0) { ERR("resolve of hostname %s failed", hostname); return -1; }
    }
    if(verbose) ERR("connecting to %{IP32}:%Hu/%s", addr, port, hostname);
    fd = lst_conn_open(addr, port, flags);
    if(fd<0) { ERR("connect to %{IP32}:%Hu/%s error %{ERR}", addr, port, hostname, errno); return -1; }
    if(verbose) ERR("connected to %{IP32}:%Hu/%s", addr, port, hostname);
    return fd;
  }
  switch(via->type) {
  default: ERR("unknown proxy type %d with address %{IP32}:%Hu", via->type, via->ip, via->port); return -1;
  case proxy_http: handler = &handle_proxy_http; name = "proxy_http"; break;
  case proxy_https: ERR0("https proxy not supported yet"); return -1;
  case proxy_socks4:
    if(!addr) { ERR0("socks4 doesn't support connecting without pre-resolving hostname to ip-address"); return -1; }
    handler = &handle_proxy_socks4; name = "socks4"; break;
  case proxy_socks5: handler = &handle_proxy_socks5; name = "socks5"; break;
  case proxy_r:
    if(!addr) { ERR0("proxy_r doesn't support connecting without pre-resolving hostname to ip-address"); return -1; }
    handler = &handle_proxy_r; name = "proxy_r"; break;
  }
  if(verbose) ERR("connecting to %s://%{IP32}:%Hu", name, via->ip, via->port);
  fd = lst_conn_open(via->ip, via->port, 0);
  if(fd<0) { ERR("connect to %s://%{IP32}:%Hu error %{ERR}", name, via->ip, via->port, errno); return -1; }
  if(verbose) ERR("connecting to %{IP32}:%Hu through %s://%{IP32}:%Hu", addr, port, name, via->ip, via->port);
  r = handler(fd, addr, port, hostname, via, logfn, cbdata);
  if(r<=0) {
    if(r==0) ERR("%s://%{IP32}:%Hu connect to %{IP32}:%Hu/%s error: proxy closed connection", name, via->ip, via->port, addr, port, hostname);
    if(r==-1) ERR("%s://%{IP32}:%Hu connect to %{IP32}:%Hu/%s error %{ERR}", name, via->ip, via->port, addr, port, hostname, errno);
    if(r==-3) ERR("%s://%{IP32}:%Hu connect to %{IP32}:%Hu/%s error: proxy response malformed", name, via->ip, via->port, addr, port, hostname);
    close(fd); return -1;
  }
  if(verbose) ERR("connected to %{IP32}:%Hu through %s://%{IP32}:%Hu", addr, port, name, via->ip, via->port);
  if((flags & LST_FLAG_NONBLOCK_AFTER) && lst_conn_set_nonblock(fd)<0) {
    ERR("set_nonblock() error %{ERR}", errno);
    close(fd); return -1;
  }
  return fd;
}

static int handle_proxy_http(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata) {
  qprint_ctx out;
  char buf[256], ipstr[50];
  size_t p;
  ssize_t r;
  uint status;
  char const *b;
  qp_init_buf(&out, buf, sizeof(buf));
  out.fd = fd;
  if(addr) { qp_snprintf(ipstr, sizeof(ipstr), "%{IP32}", addr); hostname = ipstr; }
  else if(!hostname) { ERR0("proxy_http: both target addr and target hostname missing"); return -2; }
  qp_printf(&out,"CONNECT %s:%Hu HTTP/1.1\r\nHost: %s:%Hu\r\n", hostname, port, hostname, port);
  if(via->auth) qp_printf(&out,"Proxy-Authorization: %s\r\n", via->auth);
  qp_puts(&out, "\r\n");
  qp_flush(&out);
  if(out.fail) { errno = (out.fail==-1)?0:out.fail; return -1; }
  p = 0;
  lst_conn_wait_readable(fd, 1000);
  while(1) {
    assert(p<sizeof(buf));
    r = read(fd, buf+p, 1); /* we shouldn't fetch extra bytes after proxy response */
    if(r>0) {
      assert(r==1); /* assert((size_t)r<=sizeof(buf)-p); */
      if(buf[p]=='\n') break;
      p++;
      if(p==sizeof(buf)) return -3;
    } else if(!r) return 0;
    else if(errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR) return -1;
    else lst_conn_wait_readable(fd, 1000);
  }
  buf[p] = 0;
  if(p && buf[p-1]=='\r') buf[p-1] = 0;
  b = buf+9;
  if(strncmp(buf,"HTTP/1.",7) || buf[7]!='0' && buf[7]!='1' || buf[8]!=' ' || buf[9]<'1' || buf[9]>'9' || strp_to_uint(&b,&status)<0) return -3;
  if(status<200 || status>299) {
    ERR("proxy_http://%{IP32}:%Hu connect to %{IP32}:%Hu/%s error: proxy response HTTP/%u", via->ip, via->port, addr, port, hostname, status);
    return -2;
  }
  status = 1;
  while(1) {
    r = read(fd, buf, 1);
    if(r>0) {
      assert(r==1);
      p++;
      if(buf[0]=='\r') { if(status==1) status=2; else status = 0; }
      else if(buf[0]=='\n') { if(status==1 || status==2) break; else status=1; }
      else status = 0;
      if(p>=16384) return -3;
    } else if(!r) return 0;
    else if(errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR) return -1;
    else lst_conn_wait_readable(fd, 1000);
  }
  return 1;
}

static int handle_proxy_socks4(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata) {
  int r;
  size_t p;
  uchar buf[9];
  if(!addr) { ERR0("proxy_socks4: target addr missing (socks4 doesn't support connecting directly to hostname)"); return -2; }
  buf[0] = 4;
  buf[1] = 1;
  setU16BE(buf+2, port);
  setU32BE(buf+4, addr);
  buf[8] = 0;
  if(lst_conn_send_all(fd, buf, 9)<0) return -1;
  p = 0;
  lst_conn_wait_readable(fd, 1000);
  while(p<8) {
    r = read(fd, buf+p, 8-p);
    if(r>0) {
      assert((uint)r<=8-p);
      p+=r;
      if(buf[0]!=0) return -3;
      if(p>=2 && buf[1]!=0x5A) goto rej;
      continue;
    }
    else if(!r) return 0;
    else if(errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR) return -1;
    else lst_conn_wait_readable(fd, 1000);
  }
  return 1;
rej:
  ERR("socks4://%{IP32}:%Hu connect to %{IP32}:%Hu/%s error: connection rejected", via->ip, via->port, addr, port, hostname);
  return -2;
}

static int recv_all(int fd, uchar *b, uint n) {
  uint p;
  int r;
  lst_conn_wait_readable(fd, 1000);
  for(p=0; p<n; ) {
    r = read(fd, b+p, n-p);
    if(r>0) { assert((uint)r<=n-p); p+=r; continue; }
    else if(!r) return 0;
    else if(errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR) return -1;
    else lst_conn_wait_readable(fd, 1000);
  }
  return 1;
}

static int handle_proxy_socks5(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata) {
  int r;
  size_t p;
  uchar buf[300];
  if(!addr && !hostname) { ERR0("proxy_socks5: both target addr and target hostname missing"); return -2; }

  buf[0] = 5;
  buf[1] = 1;
  buf[2] = 0;
  if(lst_conn_send_all(fd, buf, 3)<0) return -1;
  if((r=recv_all(fd,buf,2))<=0) return r;
  if(buf[0]!=5) return -3;
  if(buf[1]==0xFF) { ERR("socks5://%{IP32}:%Hu connect to %{IP32}:%Hu/%s error: auth required (we don't support it anyway)", via->ip, via->port, addr, port, hostname); return -2; }
  if(buf[1]!=0) return -3;
  
  buf[0] = 5;
  buf[1] = 1;
  buf[2] = 0;
  if(addr) {
    buf[3] = 1;
    setU32BE(buf+4, addr);
    setU16BE(buf+8, port);
    if(lst_conn_send_all(fd, buf, 10)<0) return -1;
    if((r=recv_all(fd,buf,10))<=0) return r;
  } else {
    p = strlen(hostname);
    if(p>255) { ERR0("socks5 hostname should be not more 255 characters"); return -2; }
    buf[3] = 3;
    buf[4] = p;
    bcopy(hostname, buf+5, p);
    setU16BE(buf+5+p, port);
    if(lst_conn_send_all(fd, buf, 7+p)<0) return -1;
    if((r=recv_all(fd,buf,7+p))<=0) return r;
  }
  if(buf[0]!=5) return -3;
  if(buf[1]!=0) {
    ERR("socks5://%{IP32}:%Hu connect to %{IP32}:%Hu/%s error: connection rejected", via->ip, via->port, addr, port, hostname);
    return -2;
  }
  return 1;
}

static int handle_proxy_r(int fd, uint32 addr, uint16 port, char const *hostname, proxy_desc const *via, conn_open_log_fn logfn, void *cbdata) {
  int r;
  uchar buf[7];
  if(!addr) { ERR0("proxy_r: target addr missing (proxy_r doesn't support connecting directly to hostname)"); return -2; }
  buf[0] = 0;
  *(uint32*)(buf+1) = addr;
  *(uint16*)(buf+5) = port;
  if(lst_conn_send_all(fd, buf, 7)<0) return -1;
  while(1) {
    lst_conn_wait_readable(fd, 1000);
    r = read(fd, buf, 1);
    if(r==1) break;
    if(!r) return 0;
    assert(r<0);
    if(errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR) return -1;
  }
  return 1;
}
