/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/nclient.c - main source file for fproxy-client-wrapper
 * Copyright (c) 2018-2020, 2022, 2023 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <assert.h>
#include <poll.h>
#include <openssl/evp.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <fcl/types.h>
#include <fcl/endian.h>
#define FCL_SIMPLE_NAMES
#include <fcl/strparse.h>
#include <fcl/qlog.h>
#define FCL_WITH_VA_LIST
#include <fcl/qprint.h>
#include <fcl/malloc.h>
#include <libsimpletcp/libsimpletcp.h>
#include "conn_open.h"
#include "nclient.h"

/*
 * TODO: improve protocol, allow "store and forward" for requests: client sends
 *   connection info, request blob and "push" command, and then we do the connection and sending the request
 */
typedef struct {
  uint32 ip;
  uint16 port;
  char const *path;
} t_addr;

typedef struct {
  t_addr addr;
  int    backlog;
  uint   uid;
  uint   gid;
  uint   mode;
} t_laddr;

static int parse_addr(char const *a, t_addr *r) {
  char const *v;
  if(v=skip_prefix(a,"tcp:")) return lst_parse_addrstr(v, &r->ip, &r->port);
  if(v=skip_prefix(a,"unix:")) { if(!*v) return -1; r->path=v; return 0; }
  return -1;
}

static int parse_proxy(proxy_desc *r, char const *b) {
  if(r->type) {
    if(!strcmp(b,"none")) { free_proxyspec(r); return 0; }
    return -3;
  }
  return parse_proxyspec(r,b);
}

static int listen_at(t_laddr const *la) {
  int fd;
  struct stat sb;
  if(la->addr.port) {
    fd = lst_conn_listen(la->addr.ip, la->addr.port, la->backlog, LST_FLAG_REUSE_ADDR|LST_FLAG_NONBLOCK_AFTER);
    if(fd<0) { qlog("listen(tcp:%{IP32}:%Hu) error %{ERR}", la->addr.ip, la->addr.port, errno); return -1; }
    return fd;
  }
  if(lstat(la->addr.path,&sb)>=0 && S_ISSOCK(sb.st_mode)) unlink(la->addr.path);
  fd = lst_conn_listen_unix(la->addr.path, la->backlog, LST_FLAG_REUSE_ADDR|LST_FLAG_NONBLOCK_AFTER);
  if(fd<0) { qlog("listen(unix:%S) error %{ERR}", la->addr.path, errno); return -1; }
  if((la->uid || la->gid) && lchown(la->addr.path, la->uid, la->gid)<0) qlog("lchown(%s) error %{ERR}", la->addr.path, errno);
  else if(la->mode && chmod(la->addr.path, la->mode)<0) qlog("chmod(%s) error %{ERR}", la->addr.path, errno);
  else return fd;
  close(fd); unlink(la->addr.path); return -1;
}

static char * argv0;
static size_t argv0len;
static t_laddr l;
static int lfd = -1, afd = -1, preinit_ssl, verbose;
static conn_options conn_opts;
static int multiplex_out_err, target_connected;

/* our client should prioritize reading from us against writing to us */
static void write_to_client(int fd, uchar const *b, size_t len) {
  size_t csz;
  ssize_t r;
  int e;
  while(csz=len) {
    while((ssize_t)csz<=0) csz/=2;
    while((r = write(fd, b, csz))<0) {
      e = errno;
      if(e==EWOULDBLOCK && e==EAGAIN) lst_conn_wait_writable(fd,1000);
      else if(e!=EINTR) { qlog("write-to-client error %{ERR}", e); exit(-1); }
    }
    b += r;
    len -= r;
  }
}

static void send_to_client(uchar err, void const *b, size_t len) {
  size_t csz;
  uchar hdr[3];
  if(!len) return;
  if(afd<0) { /* we are in console mode */
    write_to_client(err?2:1, b, len);
    return;
  }
  switch(multiplex_out_err) {
  case 2: /* write everything to the stream before connected; write only stdout after */
    if(target_connected && err) return;
  case 0: /* dumbly write everything to the stream */
    write_to_client(afd, b, len);
    return;
  case 3: /* multiplex as mode=1 before connected, and write plain stdout after */
    if(target_connected) { if(!err) write_to_client(afd, b, len); return; }
  case 1: /* multiplex out/err as <U8 is_err> <U16 len> <U8 data[len]> */
    while(csz=len) {
      if(csz>32768) csz = 32768;
      hdr[0] = err?1:0;
      hdr[1] = csz/256;
      hdr[2] = csz%256;
      write_to_client(afd, hdr, 3);
      write_to_client(afd, b, csz);
      b = ((char*)b) + csz;
      len -= csz;
    }
    return;
  }
}

static int qp_wcb(void * cbdata, char const * b, size_t n) {
  send_to_client(cbdata?1:0, b, n);
  return 0;
}

static void init_qp_output(void) {
  qp_out.fd = -2;
  qp_out.cb = &qp_wcb;
  qp_out.cbdata = NULL;
  qp_err.fd = -2;
  qp_err.cb = &qp_wcb;
  qp_err.cbdata = &qp_err;
}

extern void raw_fail(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  qp_puts(&qp_err, "ERROR: ");
  qp_vprintf(&qp_err, fmt, arg);
  qp_putc(&qp_err, '\n');
  va_end(arg);
  qp_flush(&qp_err);
  /* may be SSL_CTX_free() here? */
  exit(-1);
}
extern void set_process_namef(char const *fmt, ...) {
  va_list arg;
  if(!argv0) return;
  va_start(arg, fmt);
  qp_vsnprintf(argv0, argv0len, fmt, arg);
  va_end(arg);
}

static void setup_minversion(conn_options *o, unsigned short minver, int legacy_server) {
  o->mask |= 1;
  o->minver = minver;
  o->legacy_server = legacy_server;
}
#define SETSTR(A,B) do { if(A) fcl_free_c(A); A = fcl_strdup_check(B); } while(0)
static void setup_client_cert(conn_options *o, char const *certfile, char const *keyfile, int flag) {
  /* optional, usually not used */
  o->mask |= 2;
  SETSTR(o->client_cert_file, certfile);
  SETSTR(o->client_pkey_file, keyfile);
}
static void setup_trusted_ca_list(conn_options *o, char const *listfile, int enforce) {
  /* this is required to verify server cert */
  o->mask |= 4;
  SETSTR(o->calist_file, listfile);
  o->enforce_verify = enforce;
  o->verify_depth = 9;
}
static int valid_hostname(char const *hostname) {
  size_t j;
  char c;
  for(j=0; c=hostname[j]; j++) if(c<33 || c>126 || c=='<' || c=='>' || c=='\\' || c=='/' || c=='?' || c=='*' || c=='|') return 0;
  return j?1:0;
}
static void setup_target(conn_options *o, uint32 ip, uint16 port, char const *sni) {
  char addrstr[50];
  if(sni && !valid_hostname(sni)) raw_fail("malformed target SNI");
  lst_print_ip_port(addrstr, ip, port);
  o->ip = ip;
  o->port = port;
  SETSTR(o->addrstr, addrstr);
  SETSTR(o->sni, sni);
}
static void setup_resolve_domain(conn_options *o, char const *name) {
  char addrstr[50];
  if(name && *name && !valid_hostname(name)) raw_fail("malformed target domain");
  if(name && !*name) name = NULL;
  SETSTR(o->domain, name);
}
static void setup_alpn_reset(conn_options *o) {
  o->alpn_size = 0;
}
static void setup_alpn_append(conn_options *o, uchar const *proto, size_t len) {
  if(len>=100) raw_fail("ALPN protoname too long");
  if(sizeof(o->alpn)-o->alpn_size<=len) raw_fail("ALPN proto list too long");
  o->alpn[o->alpn_size++] = len;
  bcopy(proto, o->alpn+o->alpn_size, len);
  o->alpn_size += len;
}

static void handle_client_connection(void);

int main(int argc, char **argv) {
  int j, dummy, rr;
  char const *a, *v, *vv;
  pid_t fr;
  char logpre[50];

  argv0 = argv[0];
  if(argv0) argv0len = strlen(argv0)+1;
  if(!argv0len) argv0 = NULL;

  l.uid = -1; l.gid = -1; l.mode = -1;
  conn_opts.preconnected_fd = -1;
  conn_opts.use_ssl = 1;
  for(j=1; j<argc; j++) {
    a = argv[j];
    if(v=skip_prefix(a,"--listen=")) {
      if(l.addr.port || l.addr.path) { qlog("duplicate --listen= option"); return -1; }
      if(parse_addr(v, &l.addr)<0 || !l.addr.port && !l.addr.path) { qlog("bad listen addr %\"S", v); return -1; }
      continue;
    }
    if(v=skip_prefix(a,"--listen-backlog=")) {
      if(l.backlog) { qlog("duplicate --listen-backlog= option"); return -1; }
      if(str_to_sint(v, &l.backlog)<0 || !l.backlog) { qlog("bad listen-backlog %\"S", v); return -1; }
      continue;
    }
    if(v=skip_prefix(a,"--listen-uid=")) {
      if(l.uid+1) { qlog("duplicate --listen-uid= option"); return -1; }
      if(str_to_uint(v, &l.uid)<0 || !(l.uid+1)) { qlog("bad listen-uid %\"S", v); return -1; }
      continue;
    }
    if(v=skip_prefix(a,"--listen-gid=")) {
      if(l.gid+1) { qlog("duplicate --listen-gid= option"); return -1; }
      if(str_to_uint(v, &l.gid)<0 || !(l.gid+1)) { qlog("bad listen-gid %\"S", v); return -1; }
      continue;
    }
    if(v=skip_prefix(a,"--listen-mode=")) {
      if(l.mode+1) { qlog("duplicate --listen-mode= option"); return -1; }
      if(str_to_uint(v, &l.mode)<0 || !(l.mode+1)) { qlog("bad listen-mode %\"S", v); return -1; }
      continue;
    }
    if(!strcmp(a,"--preinit-ssl")) {
      if(preinit_ssl) { qlog("duplicate --preinit-ssl option"); return -1; }
      preinit_ssl = 1;
      continue;
    }
    if(!strcmp(a,"--verbose") || !strcmp(a,"-v")) {
      verbose = verbose?2:1;
      continue;
    }
    if(!strcmp(a,"--fd")) {
      if(conn_opts.preconnected_fd>=0) { qlog("duplicate --fd option"); return -1; }
      conn_opts.preconnected_fd = 3;
      continue;
    }
    if(v=skip_prefix(a,"--ssl-calist=")) {
      setup_trusted_ca_list(&conn_opts, v, 0);
      continue;
    }
    if(v=skip_prefix(a,"--ssl-client-cert=")) {
      conn_opts.mask |= 2;
      SETSTR(conn_opts.client_cert_file, v);
      continue;
    }
    if(v=skip_prefix(a,"--ssl-client-pkey=")) {
      conn_opts.mask |= 2;
      SETSTR(conn_opts.client_pkey_file, v);
      continue;
    }
    if(v=skip_prefix(a,"--proxy=")) {
      rr = parse_proxy(&conn_opts.via, v);
      if(rr==-2) { qlog("out of memory parsing --proxy option"); return -1; }
      if(rr==-3) { qlog("duplicate --proxy option"); return -1; }
      if(rr<0) { qlog("bad --proxy option"); return -1; }
      continue;
    }
    qlog("unknown option: %\"S", a); return -1;
  }
  if(!l.backlog) l.backlog = -1;
  if(!(l.uid+1)) l.uid = 0;
  if(!(l.gid+1)) l.gid = 0;
  if(!(l.mode+1)) l.mode = 0;
  if(l.addr.port || l.addr.path) {
    if(conn_opts.preconnected_fd!=-1) { qlog("--listen + --fd options are incompatible"); return -1; }
    if((lfd = listen_at(&l))<0) return -1;
  }
  if(preinit_ssl) init_ssl(&conn_opts);
  if(lfd<0) { /* console mode */
    handle_client_connection();
    return 0;
  }
  qlog_set_times(QLOG_TIMES_SEC);
  while(1) {
    while(waitpid(-1,&dummy,WNOHANG)>0);
    if(lst_conn_wait_readable(lfd, 1000)<0) { qlog("waiting for connections error %{ERR}", errno); return -1; }
    afd = accept(lfd, NULL, NULL);
    if(afd<0) continue; /* don't care EINTR/EAGAIN or something another */
    if((fr=fork())<0) qlog("fork() error %{ERR}", errno);
    else if(!fr) {
      qp_snprintf(logpre, sizeof(logpre), "[%u]", (uint)getpid());
      qlog_set_prefix(logpre);
      init_qp_output();
      handle_client_connection();
      exit(0);
    }
    close(afd);
  }
}

/************************************************************************************************************************************/
/************************************************************************************************************************************/
/************************************************************************************************************************************/


/* CLIENT = parent process who interacts with us via fd0/fd1/fd2
   SERVER = remote server accessed via OpenSSL */
static uchar  CSbuf[20000];
static size_t CSlen, CSpos;
static int    CSeof;
static uchar  CSQbuf[20000];
static size_t CSQhead, CSQtail;
static int update_CSbuf(void) {
  ssize_t r;
  if(CSlen<sizeof(CSbuf)) {
    r = read((afd>=0)?afd:0, CSbuf+CSlen, sizeof(CSbuf)-CSlen);
    if(r<0 && errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR && errno!=ENOBUFS && errno!=ENOMEM) exit(-1); /* client disconnected, nothing to do now */
    if(!r) CSeof = 1;
    if(r>0) { CSlen += r; return 1; }
  }
  return 0;
}

static void CS_drain(size_t n) {
  if(!n) return;
  assert(n<=CSlen);
  CSlen -= n;
  if(CSlen) memmove(CSbuf, CSbuf+n, CSlen);
}

static SSL * ssl;
static int fd = -1;
static int queued_in = 0;
static uint16 pending_data_bytes;
static int flush_queue;

typedef struct {
  char last_line[2048];
  int  verbose;
} conn_log;
static void write_conn_log(void *cbdata, char const *fmt, ...) {
#define cl (*(conn_log*)cbdata)
  va_list arg;
  char save;
  save = cl.last_line[0];
  if(!*fmt) return;
  va_start(arg, fmt);
  qp_vsnprintf(cl.last_line, sizeof(cl.last_line), fmt, arg);
  va_end(arg);
  if(!cl.last_line[0]) cl.last_line[0] = save;
  else if(cl.verbose) { qp_printf(&qp_err, "DEBUG: %s\n", cl.last_line); qp_flush(&qp_err); }
#undef cl
}

/* used by ssl too */
extern int init_raw_connection(conn_options *o) {
  int nfd;
  size_t j, k;
  conn_log cl;
  cl.last_line[0] = 0;
  cl.verbose = verbose;
  set_process_namef("%s(%s/%s)",o->use_ssl?"SSL":"RAW",o->addrstr,o->sni);
  if((nfd=o->preconnected_fd)>=0) o->preconnected_fd = -1; else {
    nfd = conn_open_via(o->ip, o->port, o->domain, &o->via, LST_FLAG_NONBLOCK_AFTER, &write_conn_log, &cl, verbose);
    if(nfd<0) raw_fail("%s", cl.last_line);
  }
  return nfd;
}

static void do_connect(void) {
  assert(!ssl && fd==-1);
  if(conn_opts.use_ssl) {
    ssl = start_ssl_connection(&conn_opts, &fd);
    qp_printf(&qp_err,"OK: connection to SSL(%s/%s) established\n", conn_opts.addrstr, conn_opts.sni);
  } else {
    fd = init_raw_connection(&conn_opts);
    qp_printf(&qp_err,"OK: connection to RAW(%s/%s) established\n", conn_opts.addrstr, conn_opts.sni);
  }
  qp_flush(&qp_err);
  if(afd>=0 && multiplex_out_err==3) write_to_client(afd, (uchar*)"\0\0\0", 3);
  target_connected = 1;
}

static int do_recv_from_server(int *work) {
  static unsigned char SCbuf[20000];
  int eof;
  size_t n;
  ssize_t r;
  if(ssl) {
    eof = 0;
    if(n = recv_ssl_bytes(ssl, fd, SCbuf, sizeof(SCbuf), &eof)) {
      send_to_client(0, SCbuf, n);
      if(work) *work = 1;
    }
    return eof?0:1;
  } else {
    if((r=read(fd, SCbuf, sizeof(SCbuf)))<0 && errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR && errno!=ENOBUFS && errno!=ENOMEM) raw_fail("recv-from-server error %{ERR}", errno);
    if(r<=0) return r?1:0;
    assert((size_t)r<=sizeof(SCbuf));
    send_to_client(0, SCbuf, r);
    if(work) *work = 1;
    return 1;
  }
}

static size_t send_to_server(uchar const *b, size_t n) {
  ssize_t r;
  if(ssl) {
    /* TODO: partial writes */
    send_ssl_bytes(ssl, fd, b, n);
    return n;
  } else {
    if((r=write(fd,b,n))<0) {
      if(errno!=EWOULDBLOCK && errno!=EAGAIN && errno!=EINTR && errno!=ENOBUFS && errno!=ENOMEM) raw_fail("send-to-server error %{ERR}", errno);
      return 0;
    }
    assert(r>0 && (size_t)r<=n);
    return r;
  }
}

static size_t handle_packet(uchar *b, size_t bn) {
  size_t n, n0;
  uint32 c_addr;
  uint16 c_port;
  int r;
  
  if(fd>=0 && !queued_in) raw_fail("BUG! handle_packet() called while connected w/o multiplexing");
  if(flush_queue) raw_fail("BUG! handle_packet() called while flush_queue still not completed");
  if(!bn) raw_fail("BUG! handle_packet() called with no input");
  if(pending_data_bytes) {
    if(bn>pending_data_bytes) bn = pending_data_bytes;
    n = (sizeof(CSQbuf)+CSQtail-CSQhead-1) % sizeof(CSQbuf);
    if(!n) raw_fail("CSQbuf get full");
    if(bn>n) bn = n;
    if(/* tautologic: CSQhead<CSQtail || CSQtail<=1 || */(n0=sizeof(CSQbuf)-CSQhead)>=bn) {
      bcopy(b, CSQbuf+CSQhead, bn);
    } else {
      bcopy(b, CSQbuf+CSQhead, n0);
      bcopy(b+n0, CSQbuf, bn-n0);
    }
    CSQhead = (CSQhead+bn)%sizeof(CSQbuf);
    if(bn==n) flush_queue = 1;
    return bn;
  }
  switch(*b) {
  case 0:
    if(bn<4) return 0;
    setup_minversion(&conn_opts, U16BE(b+1), b[3]);
    return 4;
  case 'V': /* humanfriendly alias for 0 */
    for(n=1; n<bn && b[n]!='\n'; n++);
    if(bn-n<1) return 0;
    if(n!=3) raw_fail("syntax error");
    setup_minversion(&conn_opts, U16BE(b+1)&0x0F0F, 0);
    return 4;
  case 1:
    for(n=1; n<bn && b[n]; n++);
    if(bn-n<2) return 0;
    n0 = ++n;
    for(; n<bn && b[n]; n++);
    if(bn-n<2) return 0;
    setup_client_cert(&conn_opts, (char*)(b+1), (char*)(b+n0), b[n+1]);
    return n+2;
  case 2:
    for(n=1; n<bn && b[n]; n++);
    if(bn-n<2) return 0;
    /* /etc/ssl/certs/ca-certificates.crt */
    setup_trusted_ca_list(&conn_opts, (char*)(b+1), b[n+1]);
    return n+2;
  case 'T': /* human-friendly alias for 2 */
    for(n=1; n<bn && b[n]!='\n'; n++);
    if(bn-n<1) return 0;
    b[n] = 0;
    setup_trusted_ca_list(&conn_opts, (char*)(b+1), 0);
    return n+1;
  case 3:
  case 4:
    if(bn<7) return 0;
    for(n=7; n<bn && b[n]; n++);
    if(bn-n<1) return 0;
    if(fd>=0) raw_fail("already connected");
    setup_target(&conn_opts, U32BE(b+1), U16BE(b+5), (char*)(b+7));
    if(*b==4) get_ssl_cert(&conn_opts);
    else if(!queued_in) do_connect();
    return n+1;
  case 'C': /* human-friendly alias for 3 */
  case 'c':
    for(n=1; n<bn && b[n]!='/'; n++);
    if(bn-n<1) return 0;
    b[n] = 0;
    if(lst_parse_addrstr((char*)(b+1), &c_addr, &c_port)!=3) raw_fail("syntax error");
    b[n] = '/';
    n0 = ++n;
    for( ; n<bn && b[n]!='\n'; n++);
    if(bn-n<1) return 0;
    b[n] = 0;
    if(fd>=0) raw_fail("already connected");
    setup_target(&conn_opts, c_addr, c_port, (char*)(b+n0));
    if(*b=='c') get_ssl_cert(&conn_opts);
    else if(!queued_in) do_connect();
    return n+1;
  case 5:
  case 'S':
  case 's':
    conn_opts.do_dump_certs = 1;
    return 1;
  case 6:
    for(n=1; n<bn && b[n]; n++);
    if(bn-n<1) return 0;
    setup_alpn_append(&conn_opts, b+1, n-1);
    return n+1;
  case 'P': /* human-friendly alias for 6 */
    for(n=1; n<bn && b[n]!='\n'; n++);
    if(bn-n<1) return 0;
    setup_alpn_append(&conn_opts, b+1, n-1);
    return n+1;
  case 7:
    if(bn<2) return 0;
    conn_opts.use_ssl = b[1] & 1;
    return 2;
  case 'N':
    for(n=1; n<bn && b[n]!='\n'; n++);
    if(bn-n<1) return 0;
    if(n!=2 || b[1]!='0' && b[1]!='1') raw_fail("syntax error");
    conn_opts.use_ssl = b[1] & 1;
    return 3;
  case 8:
    queued_in = 1;
    return 1;
  case 9:
    if(bn<3) return 0;
    if(!queued_in) raw_fail("got queue-append command when not in queued mode");
    pending_data_bytes = U16BE(b+1);
    /* bytes will be handled outside */
    return 3;
  case 10:
  case 11:
    if(!queued_in) raw_fail("got queue-flush command when not in queued mode");
    flush_queue = 1;
    if(*b==11) queued_in = 0;
    return 1;
  case 12:
    if(bn<2) return 0;
    multiplex_out_err = b[1];
    return 2;
  case 13:
    if(bn<2) return 0;
    if(b[1]==0) {
      if(conn_opts.via.auth) fcl_free_c_check(conn_opts.via.auth);
      bzero(&conn_opts.via, sizeof(conn_opts.via)); return 2;
    }
    if(conn_opts.via.type) raw_fail("proxy already specified, reset it first to override");
    if(bn<9) return 0;
    if(!b[8]) {
      conn_opts.via.type = b[1];
      conn_opts.via.ip = U32BE(b+2);
      conn_opts.via.port = U16BE(b+6);
      SETSTR(conn_opts.via.auth, NULL);
      return 9;
    }
    for(n=9; n<bn && b[n]; n++);
    if(bn-n<1) return 0;
    conn_opts.via.type = b[1];
    conn_opts.via.ip = U32BE(b+2);
    conn_opts.via.port = U16BE(b+6);
    SETSTR(conn_opts.via.auth, (char*)b+9);
    return n+1;
  case 'p':
    for(n=1; n<bn && b[n]!='\n'; n++);
    if(bn-n<1) return 0;
    b[n] = 0;
    if((r=parse_proxy(&conn_opts.via, (char*)b+1))<0) {
      if(r==-3) raw_fail("proxy already specified, reset it first to override");
      if(r==-2) raw_fail("out of memory");
      raw_fail("syntax error");
    }
    return n+1;
  case 14:
    for(n=1; n<bn && b[n]; n++);
    if(bn-n<1) return 0;
    setup_resolve_domain(&conn_opts, (char*)b+1);
    return n+1;
  case 'D':
    for(n=1; n<bn && b[n]!='\n'; n++);
    if(bn-n<1) return 0;
    b[n] = 0;
    setup_resolve_domain(&conn_opts, (char*)b+1);
    return n+1;
  default:
    raw_fail("bad opcode 0x%02X", b[0]);
    return 0;
  }
}

static void handle_client_connection(void) {
  int     need_more_input, sr;
  uchar   *b;
  size_t  n;
  struct pollfd pfd[2];
  
  if(afd>=0 && lst_conn_set_nonblock(afd)
     || afd<0 && (lst_conn_set_nonblock(0)<0 || lst_conn_set_nonblock(1)<0)
     || conn_opts.preconnected_fd>=0 && lst_conn_set_nonblock(conn_opts.preconnected_fd)<0) raw_fail("can't set nonblock");
  ssl = NULL;
  
  CSlen = 0; CSpos = 0; CSeof = 0; need_more_input = 0; sr = 0;
  while(1) { /* pre-connection loop */
    if(sr) sr = 0;
    else if(flush_queue) {
      if(fd>=0) {
        pfd[0].fd = fd;
        pfd[0].events = POLLIN|POLLOUT;
        pfd[0].revents = 0;
        poll(pfd, 1, 30000);
      }
    } else if(need_more_input) {
      pfd[0].fd = (afd>=0)?afd:0;
      pfd[0].events = POLLIN;
      pfd[0].revents = 0;
      if(fd<0) poll(pfd, 1, 30000);
      else {
        pfd[1].fd = fd;
        pfd[1].events = POLLIN;
        pfd[1].revents = 0;
        poll(pfd, 2, 30000);
      }
    }
    if(fd>=0 && !do_recv_from_server(&sr)) goto fin;
    if(flush_queue) {
      /* queue flush triggered by a command 10/11 or automatically due to CSQbuf overflow;
         we should flush the CSQ and do not receive/handle anything from the client until this done;
         but we may receive bytes from the server and send them to client;
         possible states:
           cmd10:  pending_data_bytes==0 && queued_in==1
           cmd11:  pending_data_bytes==0 && queued_in==0
           auto:   pending_data_bytes>=0 && queued_in==1
           all cases may be already connected or still not connected
           when pending_data_bytes>0, we dont touch them until queue-flush is done
         */
      assert(!pending_data_bytes || queued_in);
      if(fd<0) do_connect();
      if(CSQhead>CSQtail) n = CSQhead-CSQtail;
      else if(CSQhead-CSQtail) n=sizeof(CSQbuf)-CSQtail;
      else { flush_queue=0; n=0; }
      if(n) {
        n = send_to_server(CSQbuf+CSQtail, n);
        CSQtail = (CSQtail+n)%sizeof(CSQbuf);
      }
      if(flush_queue) continue;
    }
    if(fd>=0 && !queued_in) break;
    if(pending_data_bytes && CSpos<CSlen) {
      n = handle_packet(CSbuf+CSpos, CSlen-CSpos);
      assert(n);
      CSpos += n;
      continue;
    }
    if(CSpos) { CS_drain(CSpos); CSpos = 0; }
    if(!CSlen) need_more_input = 1;
    if(update_CSbuf()) need_more_input = 0;
    if(need_more_input) {
      if(CSeof) exit(0);
      if(CSlen==sizeof(CSbuf)) raw_fail("input data malformed");
      continue;
    }
    while(CSpos<CSlen && !(fd>=0 && !queued_in || flush_queue)) {
      if(!(n = handle_packet(CSbuf+CSpos, CSlen-CSpos))) { need_more_input = 1; break; }
      CSpos += n;
    }
    if(CSpos) { CS_drain(CSpos); CSpos = 0; }
    if(!CSlen) need_more_input = 1;
  }
  assert(!pending_data_bytes && !queued_in && fd>=0 && !flush_queue);
  
  while(1) { /* connected loop */
    if(sr) sr = 0;
    else {
      pfd[0].fd = fd;
      pfd[0].events = POLLIN|(CSlen?POLLOUT:0);
      pfd[0].revents = 0;
      if(CSlen==sizeof(CSbuf)) poll(pfd, 1, 30000);
      else {
        pfd[1].fd = (afd>=0)?afd:0;
        pfd[1].events = POLLIN;
        pfd[1].revents = 0;
        poll(pfd, 2, 30000);
      }
    }
    if(!do_recv_from_server(&sr)) break;
    update_CSbuf();
    if(CSlen) CS_drain(send_to_server(CSbuf, CSlen));
    else if(CSeof) break;
  }
fin:
  if(ssl) end_ssl_connection(ssl, fd, 0);
  else if(close(fd)<0) raw_fail("close() errno %{ERR}", errno);
}

