/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * helpers/nclient_ssl.c - SSL support for fproxy-client-wrapper
 * Copyright (c) 2018-2020, 2022, 2023 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <openssl/evp.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <fcl/types.h>
#include <fcl/endian.h>
#define FCL_SIMPLE_NAMES
#include <fcl/strparse.h>
#include <fcl/qlog.h>
#define FCL_WITH_VA_LIST
#include <fcl/qprint.h>
#include <fcl/malloc.h>
#include <libsimpletcp/libsimpletcp.h>
#include "cert_match.h"
#include "certs_pin.h"
#include "cert_verify.h"
#include "conn_open.h"
#include "nclient.h"

static SSL_CTX *ctx;
static int ssl_print_errors_cb(char const *st, size_t len, void *cbdata) {
  qp_putb(&qp_err, st, len);
  qp_putc(&qp_err, '\n');
  return 0;
}
static void ssl_fail(char const * fmt, ...) {
  va_list arg;
  va_start(arg, fmt);
  qp_puts(&qp_err, "ERROR: ");
  qp_vprintf(&qp_err, fmt, arg);
  qp_putc(&qp_err, '\n');
  va_end(arg);
  ERR_print_errors_cb(&ssl_print_errors_cb, NULL);
  qp_flush(&qp_err);
  /* may be SSL_CTX_free() here? */
  exit(-1);
}

static void apply_ssl_options(conn_options const *o);
extern void init_ssl(conn_options *o) {
  if(!ctx) {
    /**** init[1] register error strings */
    /* ERR_load_BIO_strings() - what is it? */
    SSL_load_error_strings();
    /**** init[2] register algorythms */
    SSL_library_init();
    /* OpenSSL_add_all_algorithms(); - if ALL algorythms needed, not just SSL/TLS (idk what it means, see manpages) */
    /**** init[3] create ssl context */
    ctx = SSL_CTX_new(SSLv23_client_method()); /* this construction means: create context and allow all known protocols; see manpage if doubt */
    if(!ctx) ssl_fail("SSL_CTX_new() failed");
  }
  apply_ssl_options(o);
  o->mask = 0;
}
extern void free_ssl(void) {
/* it seems that this doesn't need to be called anyway;
 * also idk if it valid to do: init_ssl(), fork() a lot, free_ssl() in each child */
  /**** done[3] free ssl context */
  SSL_CTX_free(ctx);
  /**** done[2] unregoster algorythms */
  EVP_cleanup();
  /**** done[1] unregister error strings */
  ERR_free_strings();
}
static void apply_ssl_options(conn_options const *o) {
  if(o->mask & 1) {
    if(o->minver>0x0200) SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);
    if(o->minver>0x0300) SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv3);
    if(o->minver>0x0301) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1);
#ifdef SSL_OP_NO_TLSv1_1
    if(o->minver>0x0302) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1_1);
#endif
#ifdef SSL_OP_NO_TLSv1_2
    if(o->minver>0x0303) SSL_CTX_set_options(ctx, SSL_OP_NO_TLSv1_2);
#endif
    /* allow "legacy insecure renegotiation with unpatched servers"; idk what is it; see man SSL_CTX_set_options */
    /* "legacy server" means server with openssl <0.9.8m which is vulnerable;
       enabling or disabling of this option does not help to avoid vulnerability, but may prevent from connecting to vulnerable server;
       client have no way to prevent attack if server is vulnerable */
    if(o->legacy_server) SSL_CTX_set_options(ctx, SSL_OP_LEGACY_SERVER_CONNECT);
    else SSL_CTX_clear_options(ctx, SSL_OP_LEGACY_SERVER_CONNECT);
  }
  if(o->mask & 2) {
    if(SSL_CTX_use_certificate_file(ctx, o->client_cert_file, SSL_FILETYPE_PEM)<=0) ssl_fail("failed loading client-cert file");
    if(SSL_CTX_use_PrivateKey_file(ctx, o->client_pkey_file, SSL_FILETYPE_PEM)<=0) ssl_fail("failed loading client-pkey file");
    /*if(o->client_cert_file && o->client_pkey_file) {*/
    if(!SSL_CTX_check_private_key(ctx)) raw_fail("client-cert and client-pkey mismatch");
  }
  if(o->mask & 4) {
    if(o->calist_file && *o->calist_file && !SSL_CTX_load_verify_locations(ctx, o->calist_file, NULL)) ssl_fail("error loading trusted ca list");
    SSL_CTX_set_verify(ctx, o->enforce_verify?SSL_VERIFY_PEER:SSL_VERIFY_NONE, NULL);
    /* in SSL_VERIFY_NONE mode, certificate still checked but not fatal;
       SSL_get_verify_result() can be used to get results */
    SSL_CTX_set_verify_depth(ctx, o->verify_depth); /* allow max N ca-certs (N+1 total including server-cert) in server-cert-chain */
  /*SSL_CTX_set_verify_depth
    SSL_CTX_set_cert_verify_callback() may be used to replace default cert verify function X509_verify_cert() with custom one
   */
  }
}


/************************************************************************************************************************************/
/************************************************************************************************************************************/
/************************************************************************************************************************************/




static void verify_cert(SSL * ssl, X509 * server_cert, char const * hostname) {
  int match;
  long vres;
  unsigned char * bincert;
  int bincertlen;
  int problems;
  char zfn[32];
  int rrr;
  
  bincert = NULL;
  bincertlen = i2d_X509(server_cert, &bincert);
  if(bincertlen<0) ssl_fail("server certificate looks like malformed - can't reencode it");
  if(!bincertlen) raw_fail("server certificate looks like malformed - zero-length encoding?");
  
  vres = SSL_get_verify_result(ssl);
  if(!hostname) {
    /* TODO: handle missing hostname somehow? currently there is not way to get NULL hostname here */
    if(vres==X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY) raw_fail("server certificate (%s) is not valid or not trusted, verify code %ld (%s) - most likely due to missing certificate chain", zfn, vres, X509_verify_cert_error_string(vres));
    if(vres!=X509_V_OK) raw_fail("server certificate (%s) is not valid or not trusted, verify code %ld (%s)", zfn, vres, X509_verify_cert_error_string(vres));
    raw_fail("missing hostname to check the certificate against");
  }
  match = cert_match(server_cert, hostname);
  problems = (vres!=X509_V_OK || match<0);

  rrr = cert_pin_check(bincert, bincertlen, hostname, problems, zfn);
  switch(rrr) {
  case -1:
    if(zfn[0]=='Q') raw_fail("server certificate should be manually revised (%s) for hostname '%s'", zfn, hostname);
    raw_fail("server certificate was discarded due to local blacklist (%s) for hostname '%s'", zfn, hostname);
  case 1:
    return; /* just accept */
  case 0:
    break;
  case -2:
    raw_fail("server certificate for hostname '%s' can't be checked due to BAD HOSTNAME", hostname);
  case -3:
    raw_fail("server certificate for hostname '%s' can't be checked due to INTERNAL FILESYSTEM PROBLEM", hostname);
  case -4:
    raw_fail("server certificate for hostname '%s' can't be checked due to EMPTY CERT", hostname);
  case -5:
    raw_fail("server certificate for hostname '%s' can't be checked due to INTERNAL MALLOC PROBLEM", hostname);
  default:
    raw_fail("server certificate for hostname '%s' can't be checked due to internal error %d", hostname, rrr);
  }

  if(vres==X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY) raw_fail("server certificate (%s) is not valid or not trusted, verify code %ld (%s) - most likely due to missing certificate chain", zfn, vres, X509_verify_cert_error_string(vres));
  if(vres!=X509_V_OK) raw_fail("server certificate (%s) is not valid or not trusted, verify code %ld (%s)", zfn, vres, X509_verify_cert_error_string(vres));
  if(match==CERT_MATCH_NO) raw_fail("server certificate (%s) is not for hostname '%s'", zfn, hostname);
  if(match<0) raw_fail("server certificate (%s) from hostname '%s' has some problems, code %d", zfn, hostname, match);
}

extern SSL * start_ssl_connection(conn_options *o, int * r_fd) {
  int fd, ret;
  SSL * ssl;
  X509 * server_cert;
  char * tmp;
  size_t j;
  char c;
  
  init_ssl(o);
#ifdef SSL_CLIENT_ALLOW_CERT_CACHE
  SSL_CTX_set_cert_verify_callback(ctx, &cert_verify_with_cache, NULL);
  cert_verify_set_flags(1);
#else
  SSL_CTX_set_cert_verify_callback(ctx, &cert_verify_without_cache, NULL);
#endif
  cert_verify_set_dump_chains(2);
  fd = init_raw_connection(o);
  ssl = SSL_new(ctx);
  if(!ssl) ssl_fail("SSL_new() failed");
  if(o->sni && !SSL_set_tlsext_host_name(ssl, o->sni)) ssl_fail("SSL_set_tlsext_host_name() failed");
#if OPENSSL_VERSION_NUMBER >= 0x10002000L
  if(o->alpn_size && SSL_set_alpn_protos(ssl, o->alpn, o->alpn_size)) ssl_fail("SSL_set_alpn_protos() failed");
#endif
  if(!SSL_set_fd(ssl, fd)) ssl_fail("SSL_set_fd() failed");
  while((ret=SSL_connect(ssl))<=0) { /* non-blocking side effects workaround */
    if(ret<0) switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    }
    ssl_fail("SSL_connect() returned %d", ret);
  }
  server_cert = SSL_get_peer_certificate(ssl);
  if(!server_cert) raw_fail("missing server certificate (how it can happen?)");

  verify_cert(ssl, server_cert, o->sni);

#ifdef DEBUG_SSL_CLIENT
  qp_printf(&qp_err, "DEBUG: protocol = %s\nDEBUG: cipher = %s\nDEBUG: verify_result = %ld\n", SSL_get_version(ssl), SSL_get_cipher(ssl), SSL_get_verify_result(ssl));
  if(!server_cert) qp_printf(&qp_err, "DEBUG: server_cert is MISSING\n"); else {
    /* X509_NAME_oneline() properly handles NULL in first arg so no check for that */
    tmp = X509_NAME_oneline(X509_get_subject_name(server_cert),NULL,0);
    qp_printf(&qp_err, "DEBUG: server_cert subject: %s\n", tmp);
    if(tmp) OPENSSL_free(tmp);
    tmp = X509_NAME_oneline(X509_get_issuer_name(server_cert),NULL,0);
    qp_printf(&qp_err, "DEBUG: server_cert issuer: %s\n", tmp);
    if(tmp) OPENSSL_free(tmp);
  }
  qp_flush(&qp_err);
#endif
  if(server_cert) X509_free(server_cert); // needs to be called after SSL_get_peer_certificate()
  *r_fd = fd;
  return ssl;
}

/* wait_reply means "wait for remote shutdown too";
   this is not essential by protocol, but may be needed in some cases by app */
extern void end_ssl_connection(SSL * ssl, int fd, int wait_reply) {
  int ret;
  while((ret=SSL_shutdown(ssl))<=0) { /* non-blocking side effects workaround */
    if(ret<0) switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    default: ssl_fail("SSL_shutdown() returned %d", ret);
    }
    if(!wait_reply) break;
  }
  if(close(fd)<0) raw_fail("close() errno %{ERR}", errno);
  SSL_free(ssl);
}

static void dump_cert(unsigned int p, X509 * cert) {
  unsigned int i;
  X509_NAME * xn;
  char * name;

  /* TODO: proper dump */

  if(p) qp_printf(&qp_out, "%*c", (size_t)p, ' ');
  xn = X509_get_subject_name(cert);
  name = X509_NAME_oneline(xn,NULL,0);
  qp_printf(&qp_out, "subject: %s\n", name);
  if(name) OPENSSL_free(name);

  if(p) qp_printf(&qp_out, "%*c", (size_t)p, ' ');
  xn = X509_get_issuer_name(cert);
  name = X509_NAME_oneline(xn,NULL,0);
  qp_printf(&qp_out, "issuer: %s\n", name);
  if(name) OPENSSL_free(name);
  qp_flush(&qp_out);
}

extern void get_ssl_cert(conn_options *o) {
  int fd, ret;
  SSL * ssl;
  X509 * server_cert;
  STACK_OF(X509) * chain;
  char * tmp;
  int i, j;
  char c;
  
  init_ssl(o);
#ifdef SSL_CLIENT_ALLOW_CERT_CACHE
  SSL_CTX_set_cert_verify_callback(ctx, &cert_verify_with_cache, NULL);
  cert_verify_set_flags(1);
#else
  SSL_CTX_set_cert_verify_callback(ctx, &cert_verify_without_cache, NULL);
#endif
  cert_verify_set_save_chains(1);
  cert_verify_set_dump_chains(2);
  fd = init_raw_connection(o);
  ssl = SSL_new(ctx);
  if(!ssl) ssl_fail("SSL_new() failed");
  if(o->sni && !SSL_set_tlsext_host_name(ssl, o->sni)) ssl_fail("SSL_set_tlsext_host_name() failed");
  if(!SSL_set_fd(ssl, fd)) ssl_fail("SSL_set_fd() failed");
  while((ret=SSL_connect(ssl))<=0) { /* non-blocking side effects workaround */
    if(ret<0) switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    }
    ssl_fail("SSL_connect() returned %d", ret);
  }
  cert_verify_set_save_chains(0);

  qp_printf(&qp_err,"GETCERT: pre-connection to SSL(%s/%s) established\n", o->addrstr, o->sni);
  qp_flush(&qp_err);
  /* TODO: use client bufs instead of printf */
  /* TODO: html? */
  qp_printf(&qp_out, "protocol = %s\ncipher = %s\nverify_result = %ld\n", SSL_get_version(ssl), SSL_get_cipher(ssl), SSL_get_verify_result(ssl));
  qp_flush(&qp_out);

  if(!(server_cert = SSL_get_peer_certificate(ssl))) qp_printf(&qp_out, "server_cert is MISSING\n");
  else {
    qp_printf(&qp_out,"server_cert:\n");
    dump_cert(2, server_cert);
    X509_free(server_cert); // needs to be called after SSL_get_peer_certificate()
  }
  if(!(chain = cert_verify_get_saved_untrusted())) qp_printf(&qp_out, "untrusted_chain is MISSING\n");
  else {
    qp_printf(&qp_out,"untrusted_chain:\n");
    i = sk_X509_num(chain);
    for(j=0; j<i; j++) {
      server_cert = sk_X509_value(chain, j);
      qp_printf(&qp_out,"  untrusted[%d]:\n", j);
      dump_cert(4, server_cert);
    }
    sk_X509_free(chain);
  }
  if(!(chain = cert_verify_get_saved_trusted())) qp_printf(&qp_out, "trusted_chain is MISSING\n");
  else {
    qp_printf(&qp_out,"trusted_chain:\n");
    i = sk_X509_num(chain);
    for(j=0; j<i; j++) {
      server_cert = sk_X509_value(chain, j);
      qp_printf(&qp_out,"  trusted[%d]:\n", j);
      dump_cert(4, server_cert);
    }
    sk_X509_free(chain);
  }
  qp_flush(&qp_out);
  end_ssl_connection(ssl, fd, 0);
  set_process_namef("fproxy-client-wrapper");
}

/* SSL_write() is not in-place replacement for write() in case of being deferred;
   it should be retries with exactly same args and same data in buffer;
   also it usually does success until all bytes sent;
   see SSL_set_mode SSL_MODE_ENABLE_PARTIAL_WRITE, SSL_MODE_ACCEPT_MOVING_WRITE_BUFFER for details */
/* this wrapper always sends all bytes until return or get fatal error */
extern void send_ssl_bytes(SSL * ssl, int fd, unsigned char const * b, size_t n) {
  int ret;
  int csz;
  while(n) {
    csz = 16000;
    if(((unsigned)csz)>n) csz = (int)n;
    ret = SSL_write(ssl, b, csz);
    if(ret>0) {
      if(ret>csz) raw_fail("BUG! ret>csz for SSL_write()");
      n -= ret;
      b += ret;
      continue;
    }
    switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  lst_conn_wait_readable(fd, 1000); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    case SSL_ERROR_ZERO_RETURN: raw_fail("remote peer closed SSL session while we are trying SSL_write()");
    case SSL_ERROR_SYSCALL: raw_fail("SSL_write() failed due to IO errno %{ERR}", errno);
    case SSL_ERROR_SSL: ssl_fail("SSL_write() failed in protocol handling");
    default: ssl_fail("SSL_write() failed due to unknown reason");
    }
  }
}

/* may recv less bytes than requested */
extern size_t recv_ssl_bytes(SSL * ssl, int fd, unsigned char * b, size_t n, int * p_eof) {
  int ret;
  int csz;
  while(n) {
    csz = 16000;
    if(((unsigned)csz)>n) csz = (int)n;
    ret = SSL_read(ssl, b, csz);
    if(ret>0) {
      if(ret>csz) raw_fail("BUG! ret>csz for SSL_read()");
      n -= ret;
      b += ret;
      return ret;
    }
    switch(SSL_get_error(ssl,ret)) {
    case SSL_ERROR_WANT_READ:  return 0; // lst_conn_wait_readable(fd, 1); continue;
    case SSL_ERROR_WANT_WRITE: lst_conn_wait_writable(fd, 1000); continue;
    case SSL_ERROR_ZERO_RETURN: *p_eof = 1; return 0; // do not fail, it is graceful EOF // raw_fail("remote peer closed SSL session while we are trying SSL_read()");
    case SSL_ERROR_SYSCALL: raw_fail("SSL_read() failed due to IO errno %{ERR}", errno);
    case SSL_ERROR_SSL: ssl_fail("SSL_read() failed in protocol handling");
    default: ssl_fail("SSL_read() failed due to unknown reason");
    }
  }
  return 0;
}

/*
this is not correct; it only returns number of bytes that already pulled by previous actions but not
received by SSL_read(); it seems to ignore any pending data in socket buffer */
/*static int is_readable(SSL * ssl) {
  if(SSL_pending(ssl)) return 1;
  return 0;
}*/


