/*
 * This file is part of fproxy - multipurpose http(s) proxy server
 * nslookup.c - dns resolver
 * Copyright (c) 2018-2019,2023 firk (firk@cantconnect.ru)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The origin of this software must not be misrepresented; you must
 *    not claim that you wrote the original software.
 * 4. Altered versions in any form must be plainly marked as such, and
 *    must not be misinterpreted as being the original software.
 *
 * This software is provided by the author and contributors `as is'
 * without any express or implied warranty.
 */

#include <string.h>
#include <stdlib.h>
#include <netdb.h>
#include <sys/socket.h> /* for AF_INET */
#include <assert.h>
#include <fcl/types.h>
#include <fcl/endian.h>
#include <libsimpletcp/libsimpletcp.h>
#include "nslookup.h"

#define MAX_NSLOOKUP_LEN 255 /* be in sync with remote.c */
extern uint32 remote_nslookup(uint32 proxy_addr, uint16 proxy_port, char const *d) {
  int fd, r;
  size_t len;
  char buf[4];
  len = strlen(d);
  if(len<1 || len>MAX_NSLOOKUP_LEN) return LST_INADDR_NONE;
  if((fd = lst_conn_open(proxy_addr, proxy_port, 0))<0) return LST_INADDR_NONE;
  if(lst_conn_send_all(fd, d, len+1)<0) { lst_conn_close(fd,0); return LST_INADDR_NONE; }
  len = 0;
  while(len<4) {
    r = lst_conn_recv(fd, buf+len, 4-len);
    if(r<0) { lst_conn_close(fd,0); return LST_INADDR_NONE; }
    assert((unsigned)r<=4-len);
    len += r;
  }
  lst_conn_close(fd,0);
  return U32LE(buf);
}

/* TODO: way to specifiy non-system-wide DNS server */

/* return: -1=error; 0=ok, addr stored, port MAY BE stored but may be left unchanged */
extern int nslookup(char const *hostname, uint32 *addr, uint32 via_addr, uint16 via_port) {
  struct hostent *h;
  uint32 r32;
  
  if((r32 = lst_parse_ipstr(hostname))!=LST_INADDR_NONE) { *addr = r32; return 0; }
  if(strlen(hostname)>255) return -1;
  if(!via_port) {
    h = gethostbyname(hostname);
    if(!h || h->h_addrtype!=AF_INET || !h->h_addr_list || !h->h_addr_list[0]) return -1;
    *addr = U32BE(h->h_addr_list[0]);
  } else {
    r32 = remote_nslookup(via_addr, via_port, hostname);
    if(r32==LST_INADDR_NONE) return -1;
    *addr = r32;
  }
  return 0;
}

/*
<?php
$dns_server = '195.34.32.116'; //'192.168.1.1';
//$dns_server='10.77.48.115'; // megafon
function nslookup($host) {
//return $host;
  global $dns_server;
  $res = `/usr/bin/nslookup $host $dns_server 2>&1`;
  $res = explode("\n",trim($res));
  $rr=FALSE;
  foreach($res as $st) {
    $st = trim($st);
    if(substr($st,0,9)==='Address: ') $rr=trim(substr($st,9));
  }
  if($rr===FALSE) { echo 'nslookup('.htmlspecialchars($host).') failed'; exit; }

  return $rr;
}

*/




